/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue Jun 10 18:08:19 2008 by Jeff Dalton
 * Copyright: (c) 2004 - 2008, AIAI, University of Edinburgh
 */

package ix.iplan;

import java.util.*;

import ix.icore.*;
import ix.icore.domain.*;
import ix.icore.domain.event.*;

import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;
import ix.util.xml.*;

/**
 * Analyses a Domain.
 *
 * <p>This class is based on parts of O-Plan.</p>
 */
public class DomainAnalyser {

    protected Domain domain;

    protected boolean domainHasChanged = true; // true to ensure 1st analysis

    protected boolean useConditionTypes = false;

    // Refinement -> Set of Refinement
    protected DirectedGraph refinementToDirectSuccessors;

    // Refinement -> Set of Refinement
    // This is the transitive closure of refinementToDirectSuccessors.
    protected DirectedGraph refinementToReachableRefinements;

    // The remaining graphs go from refinements via the transitive closure
    // above to all of the conditions, effects, or whatever that might
    // appear if that refinement is used to expand a node that is then
    // fully expanded down to "primitives" that can't be expanded further.

    // Refinement -> Set of PatternAssignment
    protected DirectedGraph possibleConditionsTable;

    // Refinement -> Set of PatternAssignment
    protected DirectedGraph possibleEffectsTable;

    // Refinement -> Set of PatternAssignment
    protected DirectedGraph possibleUseForEffectsTable;

    // Refinement -> Set of Constraint
    // This has the constraints that aren't conditions, effects, computes,
    // or orderings.
    protected DirectedGraph possibleConstraintsTable;

    // Refinement -> Long size of the smallest full, recursive expansion.
    protected MinFullExpandFinder minFullExpandFinder;

    // /\/: It might be better to put the conditions and effects
    // in the "possible" tables into a canonical form, with all
    // variables replaced by ??.  That would avoid having entries
    // that are the same except for having different variable names.

    /**
     * Make an analyser for a domain.
     */
    public DomainAnalyser(Domain domain) {
	Debug.expect(domain != null, "null domain");
	this.domain = domain;
	domain.checkConsistency();
	domain.addDomainListener(new DomainListener() {
	    public void refinementAdded(RefinementEvent e) {
		domainHasChanged = true;
	    }
	});
    }

    public void analyseIfNeeded() {
	if (domainHasChanged) {
	    reset();
	    analyse();
	}
    }

    protected void reset() {
	refinementToDirectSuccessors = null;
	refinementToReachableRefinements = null;
	possibleConditionsTable = null;
	possibleEffectsTable = null;
	possibleUseForEffectsTable = null;
	possibleConstraintsTable = null;
	possibleConditionsCache.clear();
	possibleEffectsCache.clear();
	possibleUseForEffectsCache.clear();
	possibleConstraintsCache.clear();
	possibleTopLevelConditionsCache.clear();
	possibleTopLevelEffectsCache.clear();
	possibleTopLevelConstraintsCache.clear();
	domainHasChanged = true;
    }

    /**
     * Construct tables of derived information.
     */
    public void analyse() {
	Debug.noteln("Analyzing", domain);
	// See if we should check condition types.
	useConditionTypes = Slip.domainUsesConditionTypes(domain);
	// From refinement to refinement.
	refinementToDirectSuccessors = buildRefinementSuccessorTable();
	// Transitive closure.
	refinementToReachableRefinements =
	    Graphs.transitiveClosure(refinementToDirectSuccessors);
	DirectedGraph reflexiveReachable =
	    Graphs.makeReflexive(refinementToReachableRefinements);
	// From refinement to reachable conditions
	FanComposer cc = new FanComposer(domain.getRefinements());
	cc.apply(reflexiveReachable);
	cc.apply(Fn.accessor(Refinement.class, "getConditions"));
	possibleConditionsTable =
	    Graphs.makeDirectedGraph(cc.getResultMap());
	// From refinement to reachable effects
	FanComposer ec = new FanComposer(domain.getRefinements());
	ec.apply(reflexiveReachable);
	ec.apply(Fn.accessor(Refinement.class, "getEffects"));
	possibleEffectsTable =
	    Graphs.makeDirectedGraph(ec.getResultMap());
	// From refinement to reachable use-for-effects
	FanComposer ufec = new FanComposer(domain.getRefinements());
	ufec.apply(reflexiveReachable);
	ufec.apply(new GetUseForEffects());
	possibleUseForEffectsTable =
	    Graphs.makeDirectedGraph(ufec.getResultMap());
	// From refinement to reachable (other) constraints
	FanComposer oc = new FanComposer(domain.getRefinements());
	oc.apply(reflexiveReachable);
	oc.apply(Fn.accessor(Refinement.class, "getOtherConstraints"));
	possibleConstraintsTable =
	    Graphs.makeDirectedGraph(oc.getResultMap());
	// From refinement to minimum expansion size.
	minFullExpandFinder = new MinFullExpandFinder();
	// That's it until the domain changes again.
	domainHasChanged = false;
    }

    private class GetUseForEffects implements Function1 {
	public Object funcall(Object a) {
	    Refinement r = (Refinement)a;
	    ListOfPatternAssignment result =
		new LinkedListOfPatternAssignment();
	    for (Constraint effect: r.getEffectConstraints()) {
		if (useConditionTypes) {
		    if (! Slip.hasCondType(effect, Slip.S_USE_FOR_EFFECTS))
			continue;
		}
		PatternAssignment pv = effect.getPatternAssignment();
		if (! Slip.isAchievableCondPattern(pv.getPattern(), domain))
		    continue;
		result.add(pv);
	    }
	    return result;
	}
    }

    /* * * Queries about possible conditions * * */

    private Map possibleConditionsCache = new WeakHashMap();
    private Map possibleTopLevelConditionsCache = new WeakHashMap();

    public Set getPossibleConditions(LList pattern) {
	Set result = (Set)possibleConditionsCache.get(pattern);
	if (result == null) {
	    result = computePossibleConditions(pattern);
	    possibleConditionsCache.put(pattern, result);
	}
	return result;
    }

    protected Set computePossibleConditions(LList pattern) {
	// Map the pattern to the refinements that can expand it
	// and then to their possible Conditions.
	FanComposer fc = new FanComposer(Lisp.list(pattern));
	fc.apply(Fn.always(refinementsExpandingPattern(pattern)));
	fc.apply(possibleConditionsTable);
	Map toConditions = fc.getResultMap(); // should have only 1 key
	Debug.expect(toConditions.size() == 1);
	return (Set)toConditions.get(pattern);
    }

    public Set getPossibleConditions(PatternAssignment pv) {
	Set result = (Set)possibleConditionsCache.get(pv);
	if (result == null) {
	    result = computePossibleConditions(pv);
	    possibleConditionsCache.put(pv, result);
	}
	return result;
    }

    protected Set computePossibleConditions(PatternAssignment pv) {
	// Map the assignment to the refinements that can satisfy it
	// and then to their possible Conditions.
	FanComposer fc = new FanComposer(Lisp.list(pv));
	fc.apply(Fn.always(refinementsForCondition(pv)));
	fc.apply(possibleConditionsTable);
	Map toConditions = fc.getResultMap(); // should have only 1 key
	Debug.expect(toConditions.size() == 1);
	return (Set)toConditions.get(pv);
    }

    public Set getPossibleTopLevelConditions(LList pattern) {
	Set result = (Set)possibleTopLevelConditionsCache.get(pattern);
	if (result == null) {
	    result = new LinkedHashSet();
	    for (Refinement r: refinementsExpandingPattern(pattern)) {
		result.addAll(r.getConditions());
	    }
	    possibleTopLevelConditionsCache.put(pattern, result);
	}
	return result;
    }

    /* * * Queries about possible effects * * */

    private Map possibleEffectsCache = new WeakHashMap();  
    private Map possibleTopLevelEffectsCache = new WeakHashMap();

    public Set getPossibleEffects(LList pattern) {
	Set result = (Set)possibleEffectsCache.get(pattern);
	if (result == null) {
	    result = computePossibleEffects(pattern);
	    possibleEffectsCache.put(pattern, result);
	}
	return result;
    }

    protected Set computePossibleEffects(LList pattern) {
	// Map the pattern to the refinements that can expand it
	// and then to their possible effects.
	FanComposer fc = new FanComposer(Lisp.list(pattern));
	fc.apply(Fn.always(refinementsExpandingPattern(pattern)));
	fc.apply(possibleEffectsTable);
	Map toEffects = fc.getResultMap(); // should have only 1 key
	Debug.expect(toEffects.size() == 1);
	return (Set)toEffects.get(pattern);
    }

    public Set getPossibleEffects(PatternAssignment pv) {
	Set result = (Set)possibleEffectsCache.get(pv);
	if (result == null) {
	    result = computePossibleEffects(pv);
	    possibleEffectsCache.put(pv, result);
	}
	return result;
    }

    protected Set computePossibleEffects(PatternAssignment pv) {
	// Map the assignment to the refinements that can satisfy it
	// and then to their possible effects.
	FanComposer fc = new FanComposer(Lisp.list(pv));
	fc.apply(Fn.always(refinementsForCondition(pv)));
	fc.apply(possibleEffectsTable);
	Map toEffects = fc.getResultMap(); // should have only 1 key
	Debug.expect(toEffects.size() == 1);
	return (Set)toEffects.get(pv);
    }

    public Set getPossibleTopLevelEffects(LList pattern) {
	Set result = (Set)possibleTopLevelEffectsCache.get(pattern);
	if (result == null) {
	    result = new LinkedHashSet();
	    for (Refinement r: refinementsExpandingPattern(pattern)) {
		result.addAll(r.getEffects());
	    }
	    possibleTopLevelEffectsCache.put(pattern, result);
	}
	return result;
    }

    /* * * Queries about possible use-for-effects * * */

    protected Map possibleUseForEffectsCache = new WeakHashMap();

    public Set getPossibleUseForEffects(LList pattern) {
	Set result = (Set)possibleUseForEffectsCache.get(pattern);
	if (result == null) {
	    result = computePossibleUseForEffects(pattern);
	    possibleUseForEffectsCache.put(pattern, result);
	}
	return result;
    }

    protected Set computePossibleUseForEffects(LList pattern) {
	// Map the pattern to the refinements that can expand it
	// and then to their possible effects.
	FanComposer fc = new FanComposer(Lisp.list(pattern));
	fc.apply(Fn.always(refinementsExpandingPattern(pattern)));
	fc.apply(possibleUseForEffectsTable);
	Map toEffects = fc.getResultMap(); // should have only 1 key
	Debug.expect(toEffects.size() == 1);
	return (Set)toEffects.get(pattern);
    }

    /* * * Queries about possible constraints * * */

    private Map possibleConstraintsCache = new WeakHashMap();
    private Map possibleTopLevelConstraintsCache = new WeakHashMap();

    public Set getPossibleConstraints(LList pattern) {
	Set result = (Set)possibleConstraintsCache.get(pattern);
	if (result == null) {
	    result = computePossibleConstraints(pattern);
	    possibleConstraintsCache.put(pattern, result);
	}
	return result;
    }

    protected Set computePossibleConstraints(LList pattern) {
	// Map the pattern to the refinements that can expand it
	// and then to their possible Constraints.
	FanComposer fc = new FanComposer(Lisp.list(pattern));
	fc.apply(Fn.always(refinementsExpandingPattern(pattern)));
	fc.apply(possibleConstraintsTable);
	Map toConstraints = fc.getResultMap(); // should have only 1 key
	Debug.expect(toConstraints.size() == 1);
	return (Set)toConstraints.get(pattern);
    }

    // /\/: We won't do get/computePossibleConstaints(PatternAssignment).

    public Set getPossibleTopLevelConstraints(LList pattern) {
	Set result = (Set)possibleTopLevelConstraintsCache.get(pattern);
	if (result == null) {
	    result = new LinkedHashSet();
	    for (Refinement r: refinementsExpandingPattern(pattern)) {
		result.addAll(r.getOtherConstraints());
	    }
	    possibleTopLevelConstraintsCache.put(pattern, result);
	}
	return result;
    }

    /* * * Queries about min full expand sizes * * */

    protected Map minFullExpandSizeCache = new WeakHashMap();

    public long getMinFullExpandSize(LList pattern) {
	Long result = (Long)minFullExpandSizeCache.get(pattern);
	if (result == null) {
	    result = minFullExpandFinder.getMinFullExpandSize(pattern);
	    minFullExpandSizeCache.put(pattern, result);
	}
	return result.longValue();
    }

    public long getMinFullExpandSize(PatternAssignment pv) {
	Long result = (Long)minFullExpandSizeCache.get(pv);
	if (result == null) {
	    result = minFullExpandFinder.getMinFullExpandSize(pv);
	    minFullExpandSizeCache.put(pv, result);
	}
	return result.longValue();
    }

    // /\/: We won't do min full expand size for pattern-assignments.

    /* * * Nodes that can't be expanded * * */

    public DirectedGraph getUnexpandableNodes() {
	Map map = new HashMap();
    refinementLoop:
	for (Refinement r: domain.getRefinements()) {
	    ListOfNodeSpec nodes = r.getNodes();
	    if (nodes == null)
		continue refinementLoop;
	    ListOfNodeSpec unexpandable = new LinkedListOfNodeSpec();
	    for (NodeSpec n: nodes) {
		if (refinementsExpandingPattern(n.getPattern()).isEmpty())
		    unexpandable.add(n);
	    }
	    if (!unexpandable.isEmpty())
		map.put(r, unexpandable);
	}
	return Graphs.makeDirectedGraph(map);
    }

    /* * * Refinement successors * * */

    protected DirectedGraph buildRefinementSuccessorTable() {
	Map map = new HashMap();
	for (Refinement r: domain.getRefinements()) {
	    map.put(r, directSuccessors(r));
	}
	return Graphs.makeDirectedGraph(map);
    }

    protected Set directSuccessors(Refinement r) {
	Set result = new HashSet();
	// From expansion
	ListOfNodeSpec nodes = r.getNodes();
	if (nodes != null) {
	    for (NodeSpec n: nodes) {
		LList pattern = n.getPattern();
		result.addAll(refinementsExpandingPattern(pattern));
	    }
	}
	// Via achieve
	for (Constraint cond: r.getConditionConstraints()) {
	    if (useConditionTypes) {
		if (! Slip.hasCondType(cond, Slip.S_ACHIEVE))
		    continue;
	    }
	    PatternAssignment pv = cond.getPatternAssignment();
	    result.addAll(refinementsForCondition(pv));
	}
	return result;
    }

    public Set<Refinement> refinementsExpandingPattern(LList pattern) {
	// /\/: A refinement's expands pattern might contain &rest,
	// but a node-spec's pattern (our parameter) won't, so the
	// refinement's pattern is used as the "pattern" when calling match.
	Set<Refinement> result = new HashSet<Refinement>();
	for (Refinement r: domain.getRefinements()) {
	    if (match(r.getPattern(), pattern))
		result.add(r);
	}
	return result;
    }

    public Set refinementsForCondition(PatternAssignment pv) {
	Set result = new HashSet();
	if (!Slip.isAchievableCondPattern(pv.getPattern(), domain))
	    return result;	// will be empty
	for (Refinement r: domain.getRefinements()) {
	    for (Constraint effect: r.getEffectConstraints()) {
		PatternAssignment eff = effect.getPatternAssignment();
		if (Slip.canBeUsedForEffect(effect, r, domain)
		      && match(eff.getPattern(), pv.getPattern())
		      && match(eff.getValue(), pv.getValue()))
		    result.add(r);
	    }
	}
	return result;
    }

    /* * * Min full expansion size * * */

    class MinFullExpandFinder {

	// We use longs so that we can use the Inf class for its INFINITY.

	Map sizes = new IdentityHashMap();

	Long START_MARK = new Long(-1);

	MinFullExpandFinder() {
	}

	long getMinFullExpandSize(LList pattern) {
	    Set s_set = refinementsExpandingPattern(pattern);
	    return minExpandSize(s_set);
	}

	long getMinFullExpandSize(PatternAssignment pv) {
	    Set s_set = refinementsForCondition(pv);
	    return minExpandSize(s_set);
	}

	Map makeMinFullExpandTable() {
	    for (Object r_i: domain.getRefinements())
		findSize((Refinement)r_i);
	    return sizes;
	}

	private void findSize(Refinement r) {
	    Long s = (Long)sizes.get(r);
	    if (s == START_MARK) {
		// Cycle
		sizes.put(r, Inf.INFINITY);
	    }
	    else if (s == null) {
		sizes.put(r, START_MARK);
		ListOfNodeSpec nodes = r.getNodes();
		if (nodes == null || nodes.isEmpty()) {
		    sizes.put(r, new Long(0));
		    return;
		}
		long size = nodes.size();
		for (NodeSpec n: nodes) {
		    size = Inf.add(size, getMinFullExpandSize(n.getPattern()));
		}
		sizes.put(r, size);
	    }
	}

	private long minExpandSize(Set s_set) {
	    // s_set is the set of refinements that might expand a node.
	    if (s_set.isEmpty())
		return 0;
	    long min = Long.MAX_VALUE;
	    for (Object r_i: s_set) {
		Refinement r = (Refinement)r_i;
		findSize(r);
		long s = ((Long)sizes.get(r)).longValue();
		if (s < min)
		    min = s;
	    }
	    return min;
	}
    }

    /* * * Utilities * * */

    public boolean haveCommonPatterns(Collection pvs1, Collection pvs2) {
	for (Iterator i = pvs1.iterator(); i.hasNext();) {
	    PatternAssignment pv1  = (PatternAssignment)i.next();
	    for (Iterator j = pvs2.iterator(); j.hasNext();) {
		PatternAssignment pv2 = (PatternAssignment)j.next();
		if (match(pv1.getPattern(), pv2.getPattern()))
		    return true;
	    }
	}
	return false;
    }

    public boolean mightSatisfy(PatternAssignment cond,
				PatternAssignment effect) {
	return match(cond.getPattern(), effect.getPattern())
	    && match(cond.getValue(), effect.getValue());
    }

    public boolean mightSatisfy(PatternAssignment cond,
				Collection effects) {
	for (Iterator i = effects.iterator(); i.hasNext();) {
	    PatternAssignment e = (PatternAssignment)i.next();
	    if (mightSatisfy(cond, e))
		return true;
	}
	return false;
    }

    /**
     * Simple pattern-matcher that treats all variables as wildcards.
     */
    public static boolean match(Object pat, Object dat) {
	if (pat == dat)
	    return true;
	else if (pat instanceof Variable) {
	    Object pVal = ((Variable)pat).getValue();
	    if (pVal == null) 
		return true;	// Var has not yet been given a value
	    else
		return match(pVal, dat);
	}
	else if (dat instanceof Variable) {
	    Object dVal = ((Variable)dat).getValue();
	    if (dVal == null)
		return true;	// Var has not yet been given a value
	    else
		return match(pat, dVal);
	}
	else if (pat instanceof ItemVar || dat instanceof ItemVar)
	    return true;
	else if (pat instanceof Cons) {
	    if (dat instanceof Cons) {
		LList p = (LList)pat;
		LList d = (LList)dat;
		while (p != Lisp.NIL && d != Lisp.NIL) {
		    Object a = p.car();
		    if (a == Matcher.REST) return true;
		    if (!match(a, d.car())) return false;
		    p = p.cdr();
		    d = d.cdr();
		}
		return p == d;	// same length?
	    }
	    else
		return false;
	}
	else
	    return pat.equals(dat);
    }

    /* * * Standalone application * * */

    /**
     * Test program.
     */
    public static void main(String[] argv) {
	Debug.off();
	Parameters.processCommandLineArguments(argv);
	String domainName = Parameters.getParameter("domain");
	Domain domain = (Domain)XML.readObject(Domain.class, domainName);
	DomainAnalyser ann = new DomainAnalyser(domain);
	ann.analyse();
	printGraph("Successor table:", ann.refinementToDirectSuccessors);
	printGraph("Reach table:", ann.refinementToReachableRefinements);
	printGraph("Possible conditions:", ann.possibleConditionsTable);
	printGraph("Possible effects:", ann.possibleEffectsTable);
	printGraph("Possible constraints:", ann.possibleConstraintsTable);
	printCollection("Unreachable refinements:",
			Graphs.minimalElements
			  (ann.refinementToDirectSuccessors));
	printGraph("Unexpandable nodes:", ann.getUnexpandableNodes());
	println("Min full expansion sizes:");
	{ Map m = ann.minFullExpandFinder.makeMinFullExpandTable();
	  for (Object r: m.keySet())
	      println("   " + r + ": " + Inf.asString((Long)m.get(r)));
	}
	if (Parameters.getBoolean("ask-for-conditions")) {
	    ann.askForConditionsLoop();
	}
	if (Parameters.getBoolean("ask-for-effects")) {
	    ann.askForEffectsLoop();
	}
    }

    void askForConditionsLoop() {
	askLoop("conditions", new Function1() {
	    public Object funcall(Object a) {
		return a instanceof LList
		    ? getPossibleConditions((LList)a)
		    : getPossibleConditions((PatternAssignment)a);
	    }
	});
    }

    void askForEffectsLoop() {
	askLoop("effects", new Function1() {
	    public Object funcall(Object a) {
		return a instanceof LList
		    ? getPossibleEffects((LList)a)
		    : getPossibleEffects((PatternAssignment)a);
	    }
	});
    }

    void askLoop(String resultDescr, Function1 f) {
	println("Looking for", resultDescr);
	println("Input can be a pattern or pattern = value.");
	println("A pattern should not have parens around it.");
	for (;;) {
	    try {
		String line = Util.askLine("Input:");
		if (line.equals("bye"))
		    return;
		Object query;
		if (line.indexOf("=") >= 0) {
		    String[] parts = Strings.breakAtFirst("=", line);
		    query = new PatternAssignment
			     (Lisp.elementsFromString(parts[0]),
			      Lisp.readFromString(parts[1]));
		}
		else {
		    query = Lisp.elementsFromString(line);
		}
		println("Query:", query);
		printCollection(resultDescr, (Collection)f.funcall(query));
	    }
	    catch (Throwable t) {
		Debug.noteException(t);
	    }
	}
    }

    static void printGraph(String label, DirectedGraph g) {
	// /\/: Assumes the vertexes are refinements.
	println(label);
	List refinements = sortCollection(g.getAllNodes());
	for (Iterator i = refinements.iterator(); i.hasNext();) {
	    Object v = i.next();
	    println("   ", v);
	    List successors = sortCollection(g.getSuccessors(v));
	    for (Iterator j = successors.iterator(); j.hasNext();) {
		println("      ", j.next());
	    }
	    println("");
	}
    }

    static void printCollection(String label, Collection c) {
	println(label);
	for (Iterator i = sortCollection(c).iterator(); i.hasNext();) {
	    println("   ", i.next());
	}
	println("");
    }

    static List sortCollection(Collection c) {
	List lis = new ArrayList(c);
	if (!lis.isEmpty() && lis.get(0) instanceof Refinement)
	    Collections.sort(lis, new Refinement.NameComparator());
	return lis;
    }

    private static void print(String s) {
	System.out.print(s);
    }

    private static void println(String s) {
	System.out.println(s);
    }

    private static void println(String s, Object o) {
	System.out.println(s + " " + o);
    }

}
